/***************************************************************************
    cZ11.c  -  driver for the Lexmark Z11 printer. See README for details.
    License is GPL

                              -------------------
    begin                : Mon Jun 12 14:07:33 CEST 2000

    copyright
    original             : (C) 2000 by Christian Kornblum
    v1.1.1 changes       : (C) 2000 by Som-Som (toning)
    v1.1.2 changes       : (C) 2000 by Som-Som (toning)
    v1.1.3 changes       : (C) 2000 by Som-Som (toning)
    v1.1.4 changes       : (C) 2000 by Erik Bernhardson (division by zero)
                           and Som-Som (fix point accuracy)
    v1.1.5 changes       : (C) 2002 by Mihael Hategan (variable page size)
                         : (C) 2002 by Ullrich Sigwanz (color true gray scaling)
                           (NTSC std-coding) )
    v2.0   changes       : (C) 2003 by ullsig
                           complete rework and refactoring of the code
                           implemented 200 and 300dpi
                           implemented 1200
                           removing obsolete debug code
                           optimizing of skipping empty lines
                           increased BW-print by factor 3
                           utilizing all 192 jets of the cartridge
    v2.0.1               : (C) 2003 by ullsig
                           using 2 buffers for std B&W, making printout
                           faster by adding empty lines to the foreward
                           linedrive at the end of a sweep
    v2.0.2               : using only 1 buffer for the optimized B&W
                           algorithm, removing obsolete variables
                           merging sweeps replaced by copying at the post sweeps
                           allowing to render finer bitcmyks for res<600dpi
    v2.0.3               : buffer pointer swap instead of buffer content swap
                           setting all procedures to static
                           new invokation type (option style)
    v2.1.0               : leave color saving to gs-resolution
                           new invokation
                           make sure that colors do no overlap for color print
                           use bit-device for b&W printing
                           the prolog file is part of the executable and does not
                           need to be installed separately

    v2.1.1               : added possibility to adjust the black cartridge
    v2.1.2               : added possibility to print cmyks as dithered gray-scale
    v2.1.3               : refined debugging
    v2.2                 : added interruption of print process via file with
                           interruptable PID
    v2.2.1               : refined interruption and added choice to eject the last
                           fed page on demand



    email
    Christian Kornblum   : kornblum@studst.fh-muenster.de
    Som-Som              : somsom@freemail.c3.hu
    Mihael Hategan       : hatemih@iit.edu (for now)
    Ullrich Sigwanz      : usigwanz@lycos.de, pohl.sigwanz@freesurf.ch

    $Id: cZ11-V2.c,v 1.8 2005/12/18 09:38:48 ullsig Exp $


 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   See a license at http://www.gnu.org/copyleft/gpl.html                 *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <malloc.h>


/* various constants */
#define Z11_PAGE_HEIGHT 6720
#define Z11_PAGE_WIDTH 4802

/* general */
#define ROW_GAP 16
#define BYTES_PER_HEADER 24

/* CMYK printing */
#define COLOR_PENS 64
#define COLOR_GAP 26            /* ? */
#define BYTES_PER_COLUMN 24
#define COLOR_BUFFERS 6
#define COLORS 3

#define BW_LINES (COLOR_PENS*3)
#define BW_LINE_DRIVE (BW_LINES*2)
#define BW_EMPTY_LINES_PRESET (((COLOR_PENS+COLOR_GAP)*(COLORS-1))*2)

/* the ghostscript color identifiers */
#define GS_K 0x1
#define GS_Y 0x2
#define GS_M 0x4
#define GS_C 0x8

#define Z11_K 0
#define Z11_Y 0
#define Z11_M 1
#define Z11_C 2

#ifndef True
#define True 1
#endif

#ifndef False
#define False 0
#endif

#ifndef Bool
#define Bool int
#endif

#ifndef Byte
#define Byte unsigned char
#endif

#define ERROR_ENTRY "[cZ11-V2]: "

#if (DEBUG+0)

#warning *** paper printouts will not work
#warning *** BMP-files can be found in "/tmp"

#include "dbg_bmp.h"

static char acProlog[1] = { 0 };

#define pipe_write(nul1,nul2,nul3,nul4)

#define VERSION "2.2-DEBUG"

#else

#define VERSION "2.2"

#include "prolog.h"
#define dbg_page_start( nul1, nul2, nul3 )
#define dbg_page_end( nul )
#define dbg_line_start( nul )
#define dbg_line_end( nul )
#define dbg_set_bitcmyk( nul1, nul2 )
#define dbg_set_bit( nul )
#define dbg_empty_line( nul )

#define pipe_write fwrite

#endif

#include "digits.h"

#define MAX_ADJUST 64
#define NEUTRAL_ADJUST (MAX_ADJUST/2)
#define TESTPAGE_ADJUST 0


/* successively checking for a pixel to be set */
static Byte rndColorCheck[COLORS] = { GS_Y | GS_K, GS_M | GS_K, GS_C | GS_K };

static Byte rndColorValue[COLORS] = { Z11_Y, Z11_M, Z11_C };

typedef struct tdRndAssign
{
  Byte wrap[COLORS];            /* which value for which pixel overflow */
  Byte sequence[COLORS];        /* color Z11_... */
} TRndAssign;


/* use just one rnd integer bewteen 0..35 to initialize the preload and
   sequence for the next print */
static TRndAssign rndAssign[36] = {
  /*  wrap         sequence of pixels */
  {{0, 1, 2}, {0, 1, 2}},
  {{1, 2, 0}, {0, 1, 2}},
  {{2, 0, 1}, {0, 1, 2}},
  {{2, 1, 0}, {0, 1, 2}},
  {{1, 0, 2}, {0, 1, 2}},
  {{0, 2, 1}, {0, 1, 2}},

  {{0, 1, 2}, {1, 2, 0}},
  {{1, 2, 0}, {1, 2, 0}},
  {{2, 0, 1}, {1, 2, 0}},
  {{2, 1, 0}, {1, 2, 0}},
  {{1, 0, 2}, {1, 2, 0}},
  {{0, 2, 1}, {1, 2, 0}},

  {{0, 1, 2}, {2, 0, 1}},
  {{1, 2, 0}, {2, 0, 1}},
  {{2, 0, 1}, {2, 0, 1}},
  {{2, 1, 0}, {2, 0, 1}},
  {{1, 0, 2}, {2, 0, 1}},
  {{0, 2, 1}, {2, 0, 1}},

  {{0, 1, 2}, {2, 1, 0}},
  {{1, 2, 0}, {2, 1, 0}},
  {{2, 0, 1}, {2, 1, 0}},
  {{2, 1, 0}, {2, 1, 0}},
  {{1, 0, 2}, {2, 1, 0}},
  {{0, 2, 1}, {2, 1, 0}},

  {{0, 1, 2}, {1, 0, 2}},
  {{1, 2, 0}, {1, 0, 2}},
  {{2, 0, 1}, {1, 0, 2}},
  {{2, 1, 0}, {1, 0, 2}},
  {{1, 0, 2}, {1, 0, 2}},
  {{0, 2, 1}, {1, 0, 2}},

  {{0, 1, 2}, {0, 2, 1}},
  {{1, 2, 0}, {0, 2, 1}},
  {{2, 0, 1}, {0, 2, 1}},
  {{2, 1, 0}, {0, 2, 1}},
  {{1, 0, 2}, {0, 2, 1}},
  {{0, 2, 1}, {0, 2, 1}}
};

typedef struct tgColorCheck
{
  Byte cmyk;                    /* cmyk mask to check against */
  Byte color;                   /* translated color value */
  Byte inc;                     /* increment to proceed to next pixel - or not to */
} TCheck;

/* predefined constants for TColorCheck */
#define CHK_0(n) {         0,      0, n }
#define CHK_K(n) {      GS_K , Z11_K, n }
#define CHK_Y(n) { GS_Y|GS_K , Z11_Y, n }
#define CHK_M(n) { GS_M|GS_K , Z11_M, n }
#define CHK_C(n) { GS_C|GS_K , Z11_C, n }




typedef struct tgLine
{
  Bool isReadNeeded;
  int nChecks;
  TCheck check[16];
} TLine;


typedef struct tgDitherBlock
{
  Bool isBW;
  int expandX;
  int compressX;
  int expandY;
  int nLines;
  TLine line[16];
} TDitherBlock;


static TDitherBlock bwBlocks[5] = {

  /******************* 1x1 *********************/
  {True, 1, 1, 1, 1, {
                      {True, 1, {CHK_K (1)}}
                      }},
  /******************* 1.5x1 *********************/
  {True, 3, 2, 1, 2, {
                      {True, 3, {CHK_K (1), CHK_0 (0), CHK_K (1)}},
                      {True, 3, {CHK_K (1), CHK_K (0), CHK_0 (1)}}
                      }},
  /******************* 2x1 *********************/
  {True, 2, 1, 1, 2, {
                      {True, 2, {CHK_K (0), CHK_0 (1)}},
                      {True, 2, {CHK_0 (0), CHK_K (1)}}
                      }},
  /******************* 2x2 *********************/
  {True, 2, 1, 2, 2, {
                      {True, 2, {CHK_K (0), CHK_0 (1)}},
                      {False, 0}
                      }},
  /******************* 3x2 *********************/
  {True, 3, 1, 2, 2, {
                      {True, 3, {CHK_K (0), CHK_0 (0), CHK_0 (1)}},
                      {False, 0}
                      }},
};


static TDitherBlock colorBlocks[5] = {
  /******************* 1.5x1 *********************/
  {False, 3, 2, 2, 4, {
                       {True, 6,
                        {CHK_C (1), CHK_C (0), CHK_M (1), CHK_M (0),
                         CHK_Y (1), CHK_Y (1)}},
                       {False, 6,
                        {CHK_M (0), CHK_Y (1), CHK_Y (1), CHK_C (1),
                         CHK_C (0), CHK_M (1)}},
                       {True, 6,
                        {CHK_M (0), CHK_Y (1), CHK_Y (1), CHK_C (1),
                         CHK_C (0), CHK_M (1)}},
                       {False, 6,
                        {CHK_C (1), CHK_C (0), CHK_M (1), CHK_M (0),
                         CHK_Y (1), CHK_Y (1)}}
                       }},
  /******************* 2x1 **********************/
  {False, 2, 1, 2, 4, {
                       {True, 4,
                        {CHK_C (0), CHK_M (1), CHK_Y (0), CHK_0 (1)}},
                       {False, 4,
                        {CHK_Y (0), CHK_0 (1), CHK_C (0), CHK_M (1)}},
                       {True, 4,
                        {CHK_M (0), CHK_C (1), CHK_0 (0), CHK_Y (1)}},
                       {False, 4, {CHK_0 (0), CHK_Y (1), CHK_M (0), CHK_C (1)}}
                       }},
  /******************* 1.5x2 *********************/
  {False, 3, 2, 2, 2, {
                       {True, 3, {CHK_M (0), CHK_Y (1), CHK_C (1)}},
                       {False, 3, {CHK_C (1), CHK_M (0), CHK_Y (1)}}
                       }},
  /******************* 2x2 *********************/
  {False, 2, 1, 2, 4, {
                       {True, 4,
                        {CHK_C (0), CHK_M (1), CHK_Y (0), CHK_0 (1)}},
                       {False, 4,
                        {CHK_Y (0), CHK_0 (1), CHK_C (0), CHK_M (1)}},
                       {True, 4,
                        {CHK_M (0), CHK_C (1), CHK_0 (0), CHK_Y (1)}},
                       {False, 4, {CHK_0 (0), CHK_Y (1), CHK_M (0), CHK_C (1)}}
                       }},
  /******************* 3x2 *********************/
  {False, 3, 1, 2, 6, {
                       {True, 6,
                        {CHK_C (0), CHK_0 (0), CHK_M (1), CHK_0 (0),
                         CHK_Y (0), CHK_0 (1)}},
                       {False, 6,
                        {CHK_0 (0), CHK_Y (0), CHK_0 (1), CHK_C (0),
                         CHK_0 (0), CHK_M (1)}},
                       {True, 6,
                        {CHK_M (0), CHK_0 (0), CHK_Y (1), CHK_0 (0),
                         CHK_C (0), CHK_0 (1)}},
                       {False, 6,
                        {CHK_0 (0), CHK_C (0), CHK_0 (1), CHK_M (0),
                         CHK_0 (0), CHK_Y (1)}},
                       {True, 6,
                        {CHK_Y (0), CHK_0 (0), CHK_C (1), CHK_0 (0),
                         CHK_M (0), CHK_0 (1)}},
                       {False, 6,
                        {CHK_0 (0), CHK_M (0), CHK_0 (1), CHK_Y (0),
                         CHK_0 (0), CHK_C (1)}}
                       }},
};

#define RANDOM_DITHERING -1
#define GRAY_DITHERING -2

#define PRINT_BITCMYK  0
#define PRINT_BITRGB   1
#define PRINT_BIT      2
#define PRINT_GRAYCMYK 3
#define PRINT_GRAYRGB  4

#define ColorRead(a)  ((a)!=PRINT_BIT)
#define ColorPrint(a) ((a)==PRINT_BITCMYK || (a)==PRINT_BITRGB)
#define BlackRead(a)  (!ColorRead(a))
#define BlackPrint(a) (!ColorPrint(a))

static TDitherBlock randomBlock =
/******************* 1x1 *********************/
{ False, 3, 3, 1, 1, {
                      {True, 1},
                      }
};

/* needed when no more data available */
static TLine emptyLine = { False, 0 };


/* the structure for the pixmaps */
typedef struct tgSweepBuffer
{
  int bytepos;
  int bitpos;
  int bufpos;                   /* this is used for the colors! */
  Bool unprinted;               /* does this buffer contain data? */
  long bytesize;
  int width;
  char *buffer;
} TSweepBuffer;

/***************************************************************************
 ***************************************************************************/

/* defining a function to get the color mask (cmyk or k) */
typedef int (*TRetriever) (char *, int);

/* defining a function to squeeze or count pixels */
typedef int (*TAnalyzer) (char *, int);



typedef struct tgDataPipe
{
  TDitherBlock d;               /* copy of the dithering */
  TLine *pLine;                 /* pointer to the processed line information */
  int readSize;                 /* relevant buffer to read */
  int redundantSize;            /* redundant buffer */

  Bool isRead;                  /* 1 when data in buffer */
  int bytesRead;                /* bytes actually read in buffer */

  int width;                    /* number of pixels in line */
  int maxWidth;                 /* max number of pixels */

  Bool isAnalyzed;              /* 0, when not analyized yet */
  Bool isEOF;                   /* 1 when bffer empty */
  char *buf;

  int gsHeight;
  int printHeight;

  int gsLines;
  int printLines;

  TRetriever pRetriever;
  TAnalyzer pAnalyzer;
} TDataPipe;


static Bool
CheckForTerminate (char *termFileName)
{
  int fd;                       /* file descriptor */

  if ((fd = open (termFileName, O_RDONLY)) >= 0)
    {
      char c;
      int pid;
      int val = 0;

      Bool isInNumber = False;
      Bool isInItem = False;
      Bool isMyPid = False;

      pid = getpid ();

      while (!isMyPid && read (fd, &c, 1) > 0)
        {
          if (c >= '0' && c <= '9')
            {
              if (!isInNumber && !isInItem)
                {
                  val = 0;
                  isInNumber = True;
                }
              isInItem = True;
              if (isInNumber)
                {
                  val = (val * 10) + (int) (unsigned char) (c - '0');
                }
            }
          else if (c == ' ' || c == '\t' || c == '\n' || c == '\r')
            {
              if (isInNumber)
                {
                  isMyPid = (pid == val);
                }
              isInNumber = False;
              isInItem = False;
            }
          else
            {
              isInNumber = False;
              isInItem = True;
            }
        }
      if (!isMyPid && isInNumber)
        isMyPid = (pid == val);

      close (fd);

      if (isMyPid)
        {
          fprintf (stderr,
                   "%s Got EXIT-Signal for PID(%d) passed via \"%s\"\n",
                   ERROR_ENTRY, pid, termFileName);
          return True;
        }
    }
  return False;
}

static Bool
filewrite (const void *buffer, size_t size, size_t count, FILE * stream,
           char *termFileName)
{
  if (termFileName && *termFileName && CheckForTerminate (termFileName))
    return True;
  else
    {
      pipe_write (buffer, size, count, stream);
      return False;
    }
}

/***************************************************************************
 * extract a nibble from a char buffer
 ***************************************************************************/
static int
GetCMYK (char *buffer, int nibble)
{
  static Byte gsDiv[2] = { 16, 1 };
  return (int) ((Byte) (buffer[nibble / 2]) / gsDiv[nibble & 1]) & 0x0f;
}


/***************************************************************************
 * extract a bit from a char buffer
 ***************************************************************************/
static int
GetBIT (char *buffer, int bit)
{
  static Byte gsDiv[8] = { 128, 64, 32, 16, 8, 4, 2, 1 };
  return ((int) ((Byte) buffer[bit / 8]) / gsDiv[bit & 0x07]) & 0x01;
}

/***************************************************************************
 retrieve the width of a cmyk buffer with the size "bytes"
 the width will be a multiple of 2
 ***************************************************************************/
static int
WidthCMYK (char *buf, int bytes)
{
  char *b;
  int w;
  w = bytes;
  b = buf + w;
  while (w && !*(--b))
    --w;
  return w * 2;
}

/***************************************************************************
 retrieve the width of a bit buffer with the size "bytes"
 the width will be a multiple of 8
 ***************************************************************************/
static int
WidthBIT (char *buf, int bytes)
{
  char *b;
  int w;
  w = bytes;
  b = buf + w;
  while (w && !*(--b))
    --w;
  return w * 8;
}

/***************************************************************************
 retrieve the width of a rgb buffer with the size "bytes"
 the width will be a multiple of 8
 the contained rgb data will be translated into bitcmyk format
 ***************************************************************************/
static int
WidthRGBtoCMYK (char *buf, int bytes)
{
  int i;
  int w = 0;
  for (i = 1; bytes > 0; --bytes, ++buf, ++i)
    {
      if ((*buf = (~(*buf << 1)) & 0xee) != 0x00)
        w = i;
    }
  return w * 2;
}


/***************************************************************************
 retrieve the width of a cmyk buffer with the size "bytes"
 and convert the contained data into a BIT format
 ***************************************************************************/
static int
WidthCMYKtoBIT (char *buf, int bytes)
{
#define BW_Y 1
#define BW_M 3
#define BW_C 5
#define BW_K (BW_Y+BW_M+BW_C)

  static int blackness[16] = { 0, BW_K,
    BW_Y, BW_K,
    BW_M, BW_K,
    BW_Y + BW_M, BW_K,
    BW_C, BW_K,
    BW_Y + BW_C, BW_K,
    BW_M + BW_C, BW_K,
    BW_K, BW_K
  };

  static int mask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
  static int div[] = { 16, 1 };
  int wrap;                     /* count the color blackness */
  int i;
  int w = 0;
  int acc = 0;
  int width;
  wrap = (int) (rand () / (RAND_MAX / BW_K)) % BW_K;
  width = bytes * 2;

  for (i = 0; i < width;)
    {
      int pix;
      pix = ((int) ((Byte) buf[i / 2]) / div[i & 0x01]) & 0x0f;
      wrap += blackness[pix];
      if (wrap >= BW_K)
        {
          wrap -= BW_K;
          acc |= mask[i % 8];
          w = i + 1;
        }
      ++i;
      if (i % 8 == 0 || i >= width)
        {
          buf[(i - 1) / 8] = (char) acc;
          acc = 0;
        }
    }
  return w;
}

/***************************************************************************
 retrieve the width of a rgb buffer with the size "bytes"
 and convert the contained data into a BIT format
 ***************************************************************************/
static int
WidthRGBtoBIT (char *buf, int bytes)
{
#define BW_B 114
#define BW_G 587
#define BW_R 299
#define BW_W (BW_R+BW_G+BW_B)

  static int whiteness[8] = { 0,
    BW_B,
    BW_G,
    BW_B + BW_G,
    BW_R,
    BW_R + BW_B,
    BW_R + BW_G,
    BW_W
  };

  static int mask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
  static int div[] = { 16, 1 };
  int wrap;                     /* count the color blackness */
  int i;
  int w = 0;
  int acc = 0;
  int width;
  wrap = (int) (rand () / (RAND_MAX / BW_W)) % BW_W;
  width = bytes * 2;

  for (i = 0; i < width;)
    {
      int pix;
      pix = ((int) ((Byte) buf[i / 2]) / div[i & 0x01]) & 0x07;
      wrap += whiteness[pix];
      if (wrap >= BW_W)
        {
          wrap -= BW_W;
        }
      else
        {
          acc |= mask[i % 8];
          w = i + 1;
        }
      ++i;
      if (i % 8 == 0 || i >= width)
        {
          buf[(i - 1) / 8] = (char) acc;
          acc = 0;
        }
    }
  return w;
}


/***************************************************************************
 * initialize the data pipe and preset all structures for fast read access
 ***************************************************************************/
static void
InitDataPipe (TDataPipe * p, TDitherBlock * dither, int gsWidth, int gsHeight,
              int z11PageHeight, int printMode)
{

  memset (p, 0, sizeof (*p));
  p->d = *dither;

  switch (printMode)
    {
    case PRINT_BIT:
      p->pAnalyzer = WidthBIT;
      p->pRetriever = GetBIT;
      break;

    case PRINT_BITCMYK:
      p->pAnalyzer = WidthCMYK;
      p->pRetriever = GetCMYK;
      break;

    case PRINT_BITRGB:
      p->pAnalyzer = WidthRGBtoCMYK;
      p->pRetriever = GetCMYK;
      break;

    case PRINT_GRAYCMYK:
      p->pAnalyzer = WidthCMYKtoBIT;
      p->pRetriever = GetBIT;
      break;

    case PRINT_GRAYRGB:
      p->pAnalyzer = WidthRGBtoBIT;
      p->pRetriever = GetBIT;
      break;

    default:
      p->pAnalyzer = 0;
      p->pRetriever = 0;
      break;

    }

  /* calculate size to read and size to discard */
  {
    int gsSize;
    int z11Size;
    int pixelsPerByte;

    /* how may pixels per byte */
    pixelsPerByte = BlackRead (printMode) ? 8 : 2;

    /* maximum reasonable read pixels to print out after dithering */
    p->maxWidth = (Z11_PAGE_WIDTH * dither->compressX) / dither->expandX;

    /* number of bytes to read, that would make sense */
    gsSize = (gsWidth + pixelsPerByte - 1) / pixelsPerByte;
    z11Size = (p->maxWidth + pixelsPerByte - 1) / pixelsPerByte;

    /* strip down to the reasonable sizes */
    if (gsSize < z11Size)
      {
        p->readSize = gsSize;
        p->redundantSize = 0;
      }
    else
      {
        p->readSize = z11Size;
        p->redundantSize = gsSize - z11Size;
      }
  }

  p->gsHeight = gsHeight;

  /* make sure that we hit the correct expansion */
  {
    int gsLines;
    int z11Lines;
    int expand;

    expand = p->d.expandY;

    gsLines = gsHeight * expand;
    z11Lines = (z11PageHeight / expand) * expand;

    if (gsLines < z11Lines)
      p->printHeight = gsLines;
    else
      p->printHeight = z11Lines;

  }

  p->buf = (char *) malloc (p->readSize);
  p->bytesRead = 0;
  p->isAnalyzed = False;
  p->gsLines = 0;
  p->printLines = 0;
  p->isEOF = False;
  p->pLine = &emptyLine;

}


/***************************************************************************
 * read a line from the data pipe
 * fDiscardContents==1 means: read but no need to analyze
 * fUnget==1 means: read, but keep in pipe also for the next read
 * returns the width of the bitcmyk line, -1 mark end of input file
 ***************************************************************************/
static int
ReadDataPipe (TDataPipe * p, FILE * in, Bool fDiscardContents, Bool fUnget)
{
  /* new buffer read prepared */
  if (!p->isRead)
    {
      p->isRead = True;
      p->width = 0;
      p->isAnalyzed = False;
      /* has there been a EOF before ? */
      if (!p->isEOF)
        {
          int i;
          Bool isEOF = False;
          p->bytesRead =
            fread ((void *) p->buf, sizeof (p->buf[0]), p->readSize, in);
          for (i = p->redundantSize; i > 0 && !(isEOF = (getc (in) == EOF));
               --i);
          p->isEOF = isEOF || (p->bytesRead < p->readSize);
        }
      else
        {
          /* uncomplete read marks end of file */
          p->bytesRead = 0;
        }
    }

  /* EOF, no bytes could be read */
  if (!p->bytesRead)
    return -1;

  /* analyze the length of the buffer */
  if (!p->isAnalyzed && !fDiscardContents)
    {
      p->width = p->pAnalyzer (p->buf, p->bytesRead);
      if (p->width > p->maxWidth)
        p->width = p->maxWidth;
      p->isAnalyzed = True;
    }

  /* prepare the read in of a new line from file if requested */
  if (!fUnget)
    {
      p->isRead = False;
    }

  /* when just reading to discard, no need to print (width=0) */
  if (fDiscardContents)
    return 0;

  return p->width;
}



/***************************************************************************
 * resets all counters to 0
 ***************************************************************************/
static void
StartPage (TDataPipe * p)
{
  p->gsLines = 0;
  p->printLines = 0;
  p->pLine = &emptyLine;
}

/***************************************************************************
 * eat all remaining page data in the input files
 ***************************************************************************/
static void
EndPage (TDataPipe * p, FILE * in)
{
  while (p->gsLines < p->gsHeight && ReadDataPipe (p, in, True, False) >= 0)
    {
      ++(p->gsLines);
    }
  p->gsLines = p->gsHeight;
  p->printLines = p->printHeight;
  p->pLine = &emptyLine;
}


/***************************************************************************
 * returns True, when there is still data to read
 ***************************************************************************/
static Bool
IsDataPipeEmpty (TDataPipe * p, FILE * in)
{
  return ReadDataPipe (p, in, True, True) < 0;
}


#define IsPageDone(p) ( (p)->printLines>=(p)->printHeight )
/***************************************************************************
 * return Ture, if there is data to be unfolded
 ***************************************************************************/
static Bool
GetNextLine (TDataPipe * p, FILE * in)
{
  p->pLine = &emptyLine;
  if (!IsPageDone (p))
    {
      p->pLine = &p->d.line[p->printLines % p->d.nLines];
      ++(p->printLines);
      if (p->pLine->isReadNeeded && p->gsLines < p->gsHeight)
        {
          ++(p->gsLines);
          if (ReadDataPipe (p, in, False, False) <= 0)
            p->pLine = &emptyLine;
        }
    }
  return p->pLine->nChecks && p->width;
}


/***************************************************************************
 * structure used for printing a line
 ***************************************************************************/
typedef struct tgPrintInfo
{
  TRetriever pRetriever;        /* pointer to function to get a pixel in CMYK */
  TLine *pLine;                 /* line information */
  Bool isRnd;
  TRndAssign rndAssign;
  Bool isRndPrintPending[COLORS];
  char *buf;                    /* pointer to the buffer */
  int cPix;                     /* pixel counter */
  int nPix;                     /* number of pixels in line */
  int cCheck;                   /* counter for the maks to check */
} TPrintInfo;

/***************************************************************************
 * return the number of pixels in a line and set the parameter r accordingly
 ***************************************************************************/
static int
StartLine (TDataPipe * p, TPrintInfo * r)
{
  r->isRnd = False;
  r->pLine = p->pLine;
  r->buf = p->buf;
  r->nPix = r->pLine->nChecks ? p->width : 0;
  r->cPix = 0;
  r->cCheck = 0;
  r->pRetriever = p->pRetriever;
  return (p->d.expandX * r->nPix) / p->d.compressX;
}

static int
StartRndLine (TDataPipe * p, TPrintInfo * r)
{
  int retVal, rndNumber;
  /* number of array entries */
  int rndSize = sizeof (rndAssign) / sizeof (rndAssign[0]);
  retVal = StartLine (p, r);
  /* assign the array entry */
  rndNumber = (int) (rand () / (RAND_MAX / rndSize)) % rndSize;
  r->rndAssign = rndAssign[rndNumber];
  r->isRnd = True;
  return retVal;
}

#define IsLineDone(r) ( (r)->cPix>=(r)->nPix || !(r)->pLine->nChecks)
/***************************************************************************
 * return the number of pixels in a line and set the parameter r accordingly
 ***************************************************************************/
static Bool
GetNextPixel (TPrintInfo * r, int *color)
{
  if (!IsLineDone (r))
    {
      /* normal read, mind the checks in the matrix */
      if (!r->isRnd)
        {
          TCheck *pCheck;

          int pix;

          pix = r->pRetriever (r->buf, r->cPix);

          pCheck = &r->pLine->check[r->cCheck];

          r->cPix += pCheck->inc;
          r->cCheck = (r->cCheck + 1) % r->pLine->nChecks;

          if (pix & pCheck->cmyk)
            {
              *color = pCheck->color;
              return True;
            }
        }
      /* random dithering, get the color information and only set every 3rd pixel */
      else
        {
          int c;

          /* do we have to read a new pixel tripple ? */
          if ((r->cPix % COLORS) == 0)
            {

              int i, j;
              int pix;
              /* no pending points */
              for (c = 0; c < COLORS; ++c)
                r->isRndPrintPending[c] = False;

              /* get all pixels events that wrap the pixels */
              /* run  through a pixel tripple */
              for (j = 0, i = r->cPix; j < COLORS && i < r->nPix; ++j, ++i)
                {
                  /* read the pixel */
                  pix = r->pRetriever (r->buf, i);
                  /* test the pixel for all colors */
                  for (c = 0; c < COLORS; ++c)
                    {
                      /* wrapping of the color counter */
                      int wrap;
                      /* check if there is a pixel */
                      wrap = r->rndAssign.wrap[c];
                      if (pix & rndColorCheck[c])
                        {
                          /* have we counted enough pixels to wrap the pixel counter to 0 again */
                          /* when a pixel was there a a wrap occurred there is a complete circle done */
                          if (++wrap >= COLORS)
                            {
                              r->isRndPrintPending[c] = True;
                              wrap -= COLORS;
                            }
                          /* write back the changed wrapping value */
                          r->rndAssign.wrap[c] = wrap;
                        }
                    }
                }
            }

          /* this part is always invoked */
          /* get the color index for this pixel position */
          c = r->rndAssign.sequence[(r->cPix) % COLORS];
          /* proceed to next pixel */
          ++(r->cPix);
          /* is the color pending for print */
          if (r->isRndPrintPending[c])
            {
              *color = rndColorValue[c];
              return True;
            }
        }
    }
  return False;
}


/***************************************************************************
 * release all buffers in the data pipe
 ***************************************************************************/
static void
FinishDataPipe (TDataPipe * p)
{
  if (p->buf)
    free (p->buf);
  memset (p, 0, sizeof (*p));
}




#define BUFFER_OVERHEAD 10000

/***************************************************************************
 * This writes a number of zeros to a string.
 ***************************************************************************/
static void
ClearBuffer (char *data, int bytes)
{
  memset ((void *) data, 0, bytes);
}                               /* ClearBuffer */

/***************************************************************************
 * Initialize a sweep buffer
 ***************************************************************************/
static void
SweepBufferInit (TSweepBuffer * SweepBuffer, int width, long bytesize)
{
  SweepBuffer->bytepos = 0;
  SweepBuffer->bitpos = 0;
  SweepBuffer->bufpos = 0;
  SweepBuffer->unprinted = False;
  SweepBuffer->bytesize = bytesize;
  SweepBuffer->width = width;
  if (!SweepBuffer->buffer)
    SweepBuffer->buffer = (char *) malloc (bytesize + BUFFER_OVERHEAD);
  ClearBuffer (SweepBuffer->buffer, bytesize + BUFFER_OVERHEAD);
}                               /* SweepBufferInit */

/***************************************************************************
 * Summ up some bytes for a single checksum-byte
 ***************************************************************************/
static char
checkbyte (char *stuff, int start, int end)
{
  int i;
  char ret = 0;
  for (i = start; i <= end; i++)
    ret += stuff[i];
  return ret;
}                               /* CheckByte */

/***************************************************************************
 * This puts an unterminated amount of any chars to "out". The first
 * byte of the "string" has to give the correct number of the following bytes.
 ***************************************************************************/
static Bool
fPutLString (FILE * out, char *data, char *termFileName)
{
  return filewrite ((void *) &data[1], 1, (unsigned char) (data[0]), out,
                    termFileName);
}                               /* fPutLString */

/***************************************************************************
 * This moves the paper by a defined number of lines 1200lpi!.
 ***************************************************************************/
static Bool
LexMove (FILE * out, long int pixel, char *termFileName)
{
  char command[] = { 8,
    0x1b, 0x23, 0x80, 0x0a, 0x58, 0x00, 0x00, 0x05,
    0x1b, 0x31, 0x0c, 0x02, 0x7f, 0x00, 0x00, 0xbe
  };
  command[5] = (char) (pixel);
  command[4] = (char) (pixel >> 8);
  command[8] = checkbyte (command, 2, 7);
  return fPutLString (out, command, termFileName);
}                               /* LexMove */

/***************************************************************************
 * This initializes the printer.
 ***************************************************************************/
static Bool
LexInit (FILE * out, char *termFileName)
{
  return filewrite ((void *) acProlog, 1, sizeof (acProlog), out,
                    termFileName);
}                               /* LexInit */

/***************************************************************************
 * This tells the printer to throw out his current page.
 ***************************************************************************/
static Bool
LexEOP (FILE * out, char *termFileName)
{
  char command[] = { 24,
    0x1B, 0x22, 0x80, 0x3f, 0x84, 0x00, 0x00, 0x65,
    0x1b, 0x33, 0x00, 0x00, 0x00, 0x00, 0x00, 0x33,
    0x1b, 0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x62
  };
  return fPutLString (out, command, termFileName);
}

/***************************************************************************
 * Make bytes to use both inkjet columns on the cartridge. Return 1 if
 * they are not empty.
 ***************************************************************************/
static int
MakePacketByte (char *buffer, int bytesize, int index, char *packet)
{
  int rowindex;
  *packet = 0;

  // calculate position of data byte for second row
  rowindex = index - (BYTES_PER_COLUMN * ROW_GAP);

  // set up byte
  if (index < bytesize)
    *packet = buffer[index] & 0x55;
  if ((rowindex >= 0) && (rowindex < bytesize))
    *packet |= buffer[rowindex] & 0xAA;

  return *packet;

}                               /* MakePacketBytes */

/***************************************************************************
 * Make packets to reduce the bytes sent to the printer. This is done
 * with a 24 bit mask and up to 24 bytes for the pixels.
 * The 16 pixel space between the two columns on the cartridge has
 * to be taken into account, too.
 ***************************************************************************/
static void
MakePackets (char *buffer, int colstart, int *width,
             char *packetbuffer, int *bytesize)
{
  int packbufpos = 0;
  int lastpackheadpos = 0;
  char mergebit, packet;
  char *columnbuffer;
  register int i, j;

  /* less calculating */
  columnbuffer = &buffer[colstart];

  /* do things reverse */
  for (j = *width + ROW_GAP - 1; j >= 0; --j)
    {
      lastpackheadpos = packbufpos;
      packbufpos += 4;
      packetbuffer[lastpackheadpos] = 0x80;
      packetbuffer[lastpackheadpos + 1] = 0xff;
      packetbuffer[lastpackheadpos + 2] = 0xff;
      packetbuffer[lastpackheadpos + 3] = 0xff;

      for (i = 0; i < BYTES_PER_COLUMN; i++)
        {
          /* The gap is crap. Just to save a uP... a shame. */
          if (MakePacketByte
              (columnbuffer, *bytesize, i + j * BYTES_PER_COLUMN, &packet))
            {
              /* shift the bit to the packet position and activate it */
              mergebit = 0x80 >> (i % 8);
              packetbuffer[lastpackheadpos + i / 8 + 1] ^= mergebit;

              /* add byte to the buffer */
              packetbuffer[packbufpos] = packet;
              packbufpos++;
            }
        }
    }
  /* new bytesize */
  *bytesize = packbufpos;
}                               /* MakePackets */


/***************************************************************************
 * clear the contents of a sweep
 ***************************************************************************/
static void
ClearSweep (TSweepBuffer * sweep)
{
  sweep->unprinted = False;
  ClearBuffer (sweep->buffer, sweep->bytesize + BUFFER_OVERHEAD);
}

/***************************************************************************
 * This sends a complete sweep to the printer. Black or color, no difference.
 ***************************************************************************/
static Bool
PrintSweep (TSweepBuffer * sweep, FILE * out, int lineDrive, int *emptyLines,
            char *termFileName)
{
  char lHeader[BYTES_PER_HEADER] =
    { 0x1b, 0x40, 0x80, 0x12, 0xd2, 0x04, 0xb7, 0xFF,
    0x1b, 0x42, 0x0c, 0x04, 0x8f, 0x00, 0xFF, 0xFF,
    0x1b, 0x43, 0x00, 0xFF, 0xFF, 0x39, 0x87, 0xFF
  };

  int bytesize;
  int width;

  bytesize = sweep->bytesize;
  width = sweep->width;

  Bool will_exit = False;

  if (sweep->unprinted)
    {                           /* do not print empty sweeps */

      int bstart = 0;


      char *packetbuffer;
      int packetbuffersize;

      /* Allocate Packetbuffer */
      packetbuffersize =
        (Z11_PAGE_WIDTH + 2 * ROW_GAP) * (BYTES_PER_COLUMN + 4) + 20000;
      packetbuffer = (char *) malloc (packetbuffersize);
      memset (packetbuffer, 0, packetbuffersize);

      /* kill the noise at the end of the lines... probably MY bug */
      width += ROW_GAP;
      bytesize += BYTES_PER_COLUMN * ROW_GAP;

      /* Make the packets for the printer. */
      MakePackets (sweep->buffer, bstart, &width, packetbuffer, &bytesize);

      /* Calculate the number of bytes for the size of the pixelblock */
      lHeader[20] = (char) bytesize;
      lHeader[19] = (char) (bytesize >> 8);
      lHeader[18] = (char) (bytesize >> 16);

      /* The number of columns */
      lHeader[4] = (char) width;
      lHeader[3] = (char) (width >> 8);

      /* vertical move, always just enough for the next color */
      /* twice because once is for for 1200 lpi */
      lHeader[14] = (char) (lineDrive);
      lHeader[13] = (char) (lineDrive >> 8);

      /* checksum-bytes - this did cost me some nerve! */
      lHeader[7] = checkbyte (lHeader, 1, 6);
      lHeader[15] = checkbyte (lHeader, 9, 14);
      lHeader[23] = checkbyte (lHeader, 17, 22);

      if (*emptyLines)
        will_exit = LexMove (out, *emptyLines, termFileName);

      *emptyLines = 0;
      /* send header and compressed matrix to the printer */

      if (!will_exit)
        will_exit =
          filewrite ((void *) lHeader, 1, BYTES_PER_HEADER, out,
                     termFileName);
      if (!will_exit)
        will_exit =
          filewrite ((void *) packetbuffer, 1, bytesize, out, termFileName);

      free (packetbuffer);
      ClearSweep (sweep);
    }
  else
    *emptyLines += lineDrive;
  return will_exit;
}                               /* PrintSweep */

/***************************************************************************
 * This sends a complete sweep to the printer. Black or color, no difference.
 * but the processing is optimized for BW printing line feeds
 ***************************************************************************/
static Bool
PrintBWSweep (TSweepBuffer * sweep, FILE * out, int lineDrive,
              int *emptyLines, char *termFileName)
{
  int lEmptyLines = 0;
  int cEmptyLines;
  Bool will_exit = False;
  /* get the value only when pointer is valid */
  cEmptyLines = emptyLines ? *emptyLines : 0;
  will_exit = PrintSweep (sweep, out, lineDrive + cEmptyLines, &lEmptyLines,
                          termFileName);
  if (lEmptyLines && lineDrive + cEmptyLines == lEmptyLines && !will_exit)
    will_exit = LexMove (out, cEmptyLines, termFileName);
  /* if there is a valid pointer set it target to 0 */
  if (emptyLines)
    *emptyLines = 0;
  return will_exit;
}

/***************************************************************************
 * copy the data form src to dest (switch the buffer pointers)
 ***************************************************************************/
static void
CopySweep (TSweepBuffer * dest, TSweepBuffer * src)
{
  /* switch buffers instead of copying, faster */
  char *swapB;
  Bool swapU;

  /* swap the buffers */
  swapB = src->buffer;
  src->buffer = dest->buffer;
  dest->buffer = swapB;

  /* swap the unprinted flag */
  swapU = src->unprinted;
  src->unprinted = dest->unprinted;
  dest->unprinted = swapU;
}



/***************************************************************************
 * set a color pixel for the bit position pos in the color array
 ***************************************************************************/
static void
SetSweepPixel (TSweepBuffer * sweeps, int color, int pos)
{
  TSweepBuffer *pCurBuffer;
  pCurBuffer = &(sweeps[sweeps[color].bufpos]);
  if (pos < pCurBuffer->width)
    {
      pCurBuffer->buffer[(pos * BYTES_PER_COLUMN) + sweeps[color].bytepos]
        |= 0x01 << sweeps[color].bitpos;
      pCurBuffer->unprinted = True;
    }
}


/***************************************************************************
 * initializes the settings of the buffer, depending on the height of the
 * sweep and on the color that the sweep refers to
 ***************************************************************************/
static void
SetSweepParams (TSweepBuffer * sweep, int color, int height)
{
  int of = color * (COLOR_GAP + COLOR_PENS) + height;
  sweep->bitpos = 7 - (of % 8);
  sweep->bytepos = 8 * ((COLORS - 1) - color) + (of / 8) % 8;
  sweep->bufpos = (of / COLOR_PENS) % COLORS;
}

/***************************************************************************
 * initializes the settings of the buffer depending on the height of the
 * sweep for black and white
 ***************************************************************************/
static void
SetBWSweepParams (TSweepBuffer * sweep, int height)
{
  sweep->bitpos = 7 - (height % 8);
  sweep->bytepos = (height / 8) % BYTES_PER_COLUMN;
  sweep->bufpos = 0;
}


/***************************************************************************
* structure defining the invocation of the print routines
****************************************************************************/
typedef struct tgParameters
{
  int printMode;

  int dithering;
  int pageWidth;
  int pageHeight;

  int adjustment;
  Bool ejectOnTerminate;
  char termFileName[256];
} TParameters;


/***************************************************************************
 * This is the main printing routine. Wicked and insane. Nonetheless working.
 * This one is printing CMYK and expects the ghostscript "bitcmyk" device.
 ***************************************************************************/
static int
LexPrintCMYK (FILE * in, FILE * out, TParameters * pParam)
{
  /* buffers to hold the print information */
  TSweepBuffer sweepBuffer[2][COLOR_BUFFERS] = { {{0}} };

  /* number of pixels to move after a sweep has been printed */
  int lineDrive[2] = { 1, COLOR_PENS * 2 - 1 }; /* color printing with 1200 lpi */
  int emptyLine = 1;
  Bool isHiDither = True;
  /* struct to organize the read data */
  TDataPipe dataPipe;

  /* store dither information */
  int numPages = 0;

  /* how many buffers and sets for printing */
  int nBuffers = 2;

  Bool isRandom;
  Bool eject = False;

  isRandom = (pParam->dithering == RANDOM_DITHERING);

  /* this settings will refer when 600 lpi is used */
  if (pParam->dithering != 0 && pParam->dithering != 1)
    {
      isHiDither = False;
      nBuffers = 1;
      emptyLine = 2;
      lineDrive[0] += lineDrive[1];
      lineDrive[1] = 0;

    }

  /* initialize the data pipe with the values and allocate the memory for the buffer */
  InitDataPipe (&dataPipe,
                isRandom ? &randomBlock : &colorBlocks[pParam->dithering],
                pParam->pageWidth, pParam->pageHeight,
                Z11_PAGE_HEIGHT * nBuffers, pParam->printMode);

  /* As long as we get input... */
  while (!IsDataPipeEmpty (&dataPipe, in))
    {

      /* Reset all variables */
      int CurrentColBuffer = 0;
      int curHeight = 0;
      int cEmptyLines = 0;


      /* Initialize the printer, load a page  */
      if (LexInit (out, pParam->termFileName))
        goto LexPrintCMYK_cancelled;


      {
        int b;
        for (b = 0; b < nBuffers; ++b)
          {
            /* Get memory for the buffers (if not there yet) and clear it. */
            int i;
            for (i = 0; i < COLOR_BUFFERS; ++i)
              {
                SweepBufferInit (&sweepBuffer[b][i], Z11_PAGE_WIDTH,
                                 Z11_PAGE_WIDTH * BYTES_PER_COLUMN);
                sweepBuffer[b][i].bufpos = i;
              }
          }
      }


      StartPage (&dataPipe);

      dbg_page_start (dataPipe.printHeight, numPages, isHiDither);

      /* ... we do the pages. */
      while (!IsPageDone (&dataPipe))
        {

          Bool isLine = False;
          Bool isUnprinted = False;

          isLine = GetNextLine (&dataPipe, in);

          if (!isLine)
            dbg_empty_line (0);

          {
            int b;
            for (b = 0; !isUnprinted && b < nBuffers; ++b)
              isUnprinted = (sweepBuffer[b][0].unprinted
                             | sweepBuffer[b][1].unprinted
                             | sweepBuffer[b][3].unprinted);
          }

          /* optimize for empty lines, if buffers are empty */
          if (curHeight == 0 && !isLine && !isUnprinted)
            {
              cEmptyLines += emptyLine;
            }
          else                  /* This line does not seem to be empty or there is still data */
            {
              if (isLine)
                {
                  int b;
                  TPrintInfo r;
                  int i;

                  b = curHeight % nBuffers;

                  for (i = 0; i < COLORS; ++i)
                    SetSweepParams (&sweepBuffer[b][i], i,
                                    curHeight / nBuffers);
                  /* if data would overwrite the current buffer, store it to the overflow buffers */
                  if (sweepBuffer[b][COLORS - 1].bufpos ==
                      sweepBuffer[b][0].bufpos)
                    sweepBuffer[b][COLORS - 1].bufpos += COLORS;


                  /* start a new line and assign different dithering strategy based on dither setting */
                  if (isRandom)
                    StartRndLine (&dataPipe, &r);
                  else
                    StartLine (&dataPipe, &r);

                  dbg_line_start (0);

                  /* walk through the bitcmyk-line */
                  for (i = 0; !IsLineDone (&r); ++i)
                    {
                      int color;
                      if (GetNextPixel (&r, &color))
                        {
                          dbg_set_bitcmyk (color, i);
                          SetSweepPixel (sweepBuffer[b], color, i);
                        }
                    }

                  dbg_line_end (0);
                }

              curHeight = (curHeight + 1) % (COLORS * COLOR_PENS * nBuffers);

              /* Buffer is full or page is over. Print it. Color first... */
              if ((curHeight % (COLOR_PENS * nBuffers)) == 0
                  || IsPageDone (&dataPipe))
                {
                  int b;
                  for (b = 0; b < nBuffers; ++b)
                    {
                      if (PrintSweep (&sweepBuffer[b][CurrentColBuffer], out,
                                      lineDrive[b], &cEmptyLines,
                                      pParam->termFileName))
                        {
                          eject = pParam->ejectOnTerminate;
                          goto LexPrintCMYK_cancelled;
                        }

                      /* now handle the cyan stuff */
                      if (sweepBuffer[b][CurrentColBuffer + COLORS].unprinted)
                        CopySweep (&sweepBuffer[b][CurrentColBuffer],
                                   &sweepBuffer[b][CurrentColBuffer +
                                                   COLORS]);
                    }
                  /* switch to the next buffer */
                  CurrentColBuffer = (CurrentColBuffer + 1) % COLORS;

                }
            }


          /* this page has found an end */
          if (IsPageDone (&dataPipe))
            {
              int c;
              for (c = 0; c < COLOR_BUFFERS; c++)       /* print all unprinted buffers */
                {
                  int b;
                  Bool isUnprinted = False;
                  int BufferToPrint;
                  BufferToPrint = (c + CurrentColBuffer) % COLORS;

                  /* print both buffers, when at least one holds pixels */
                  for (b = 0; b < nBuffers; ++b)
                    isUnprinted |= sweepBuffer[b][BufferToPrint].unprinted;

                  for (b = 0; isUnprinted && b < nBuffers; ++b)
                    {
                      if (PrintSweep (&sweepBuffer[b][BufferToPrint], out,
                                      lineDrive[b], &cEmptyLines,
                                      pParam->termFileName))
                        {
                          eject = pParam->ejectOnTerminate;
                          goto LexPrintCMYK_cancelled;
                        }
                      /* now handle the cyan stuff */
                      if (sweepBuffer[b][BufferToPrint + COLORS].unprinted)
                        CopySweep (&sweepBuffer[b][BufferToPrint],
                                   &sweepBuffer[b][BufferToPrint + COLORS]);
                    }
                }
            }
        }



      /* eat any remaining whitespace as long there is data in the pipe
         so process will not hang */
      EndPage (&dataPipe, in);
      /* hand out the page */
      if (LexEOP (out, pParam->termFileName))
        {
          eject = pParam->ejectOnTerminate;
          goto LexPrintCMYK_cancelled;
        }
      /* count the pages and free memory */
      numPages++;
    }

  dbg_page_end (0);

  /* free the buffers if allocated */
LexPrintCMYK_cancelled:

  if (eject)
    {
      LexEOP (out, 0);
      dbg_page_end (0);
    }


  {
    int b;
    for (b = 0; b < nBuffers; ++b)
      {
        int i;
        for (i = 0; i < COLOR_BUFFERS; ++i)
          if (sweepBuffer[b][i].buffer)
            free (sweepBuffer[b][i].buffer);
      }
  }

  FinishDataPipe (&dataPipe);

/*
  if (numPages == 0)
    {
      fprintf (stderr, "%s No pages printed!\n\n", ERROR_ENTRY);
      return 1;
    }
*/
  return 0;
}                               /* LexPrintCMYK */



/***************************************************************************
 * return a horizontal offset, based on a global adjustment value and
 * the line that is referred
 ***************************************************************************/
static int
GetAdjustment (int adjust, int line)
{
  if (adjust < 0)
    adjust = 1;

  if (adjust > MAX_ADJUST)
    adjust = MAX_ADJUST;

  adjust -= NEUTRAL_ADJUST;
  if (adjust < 0)
    {
      adjust = -adjust;
      line = BW_LINES - 1 - line;
    }
  return (line * adjust) / BW_LINES;

}

/***************************************************************************
 * This is the main printing routine for BIT device
 * This one is printing K and expects the ghostscript "bit" device.
 ***************************************************************************/
static int
LexPrintK (FILE * in, FILE * out, TParameters * pParam)
{
  /* buffers to hold the print information */
  TSweepBuffer sweepBuffer = { 0 };

  /* struct to organize the read data */
  TDataPipe dataPipe;

  /* printed pages */
  int numPages = 0;

  Bool eject = False;

  /* initialize the data pipe with the values and allocate the memory for the buffer */
  InitDataPipe (&dataPipe, &bwBlocks[pParam->dithering],
                pParam->pageWidth, pParam->pageHeight, Z11_PAGE_HEIGHT,
                pParam->printMode);


  /* As long as we get input... */
  while (!IsDataPipeEmpty (&dataPipe, in))
    {

      int curHeight = 0;
      int cEmptyLines = 0;

      /* experimentally determined line feed on B&W */
      cEmptyLines = BW_EMPTY_LINES_PRESET;

      /* Initialize the printer, load a page  */
      if (LexInit (out, pParam->termFileName))
        goto LexPrintK_cancelled;

      SweepBufferInit (&sweepBuffer, Z11_PAGE_WIDTH,
                       Z11_PAGE_WIDTH * BYTES_PER_COLUMN);
      sweepBuffer.bufpos = 0;

      StartPage (&dataPipe);


      dbg_page_start (dataPipe.printHeight, numPages, False);

      /* ... we do the pages. */
      while (!IsPageDone (&dataPipe))
        {

          Bool isLine = False;
          /* can we just count the empty lines , or do we have to wait
             for the buffers to be printed */

          isLine = GetNextLine (&dataPipe, in);

          if (!isLine)
            dbg_empty_line (0);

          /* optimize for empty lines, if buffers are empty */
          if (curHeight == 0 && !isLine)
            cEmptyLines += 2;

          else                  /* This line does not seem to be empty or there is still data */
            {
        /******************************************************************
         * optimzed for std res black and white printing
         ******************************************************************/
              /* wait to print a complete sweep, until the first pixel of a new
                 sweep is detected (at curHeight=0)
                 when the first pixel would be set, print the old sweep and add the
                 line feeds at the end of the old sweep, which is most effcicient */

              Bool isWaitingForPixel;
              isWaitingForPixel = (curHeight == 0);

              if (isLine)
                {

                  TPrintInfo r;
                  int i;

                  /* address the lines linearily through the cartridge */
                  SetBWSweepParams (&sweepBuffer, curHeight);

                  StartLine (&dataPipe, &r);

                  dbg_line_start (0);
                  /* walk through the bit-line */
                  for (i = GetAdjustment (pParam->adjustment, curHeight);
                       !IsLineDone (&r); ++i)
                    {
                      int color;
                      if (GetNextPixel (&r, &color))
                        {
                          if (isWaitingForPixel)
                            {
                              if (PrintBWSweep
                                  (&sweepBuffer, out, BW_LINE_DRIVE,
                                   &cEmptyLines, pParam->termFileName))
                                {
                                  eject = pParam->ejectOnTerminate;
                                  goto LexPrintK_cancelled;
                                }
                              isWaitingForPixel = False;
                            }
                          dbg_set_bit (i);
                          SetSweepPixel (&sweepBuffer, 0, i);
                        }
                    }

                  dbg_line_end (0);
                }


              /* analyze the buffer state */
              if (isWaitingForPixel)
                /* new sweep line is empty, so wait with the old sweep */
                cEmptyLines += 2;
              else
                /* another line in the buffer, and wrap when the number of jets is reached */
                curHeight = (curHeight + 1) % BW_LINES;
            }


          /* this page has found an end */
          if (IsPageDone (&dataPipe))
            if (PrintBWSweep (&sweepBuffer, out, 0, (void *) 0,
                              pParam->termFileName))
              {
                eject = pParam->ejectOnTerminate;
                goto LexPrintK_cancelled;
              }
        }


      /* eat any remaining whitespace as long there is data in the pipe
         so process will not hang */
      EndPage (&dataPipe, in);
      /* hand out the page */
      if (LexEOP (out, pParam->termFileName))
        {
          eject = pParam->ejectOnTerminate;
          goto LexPrintK_cancelled;
        }

      dbg_page_end (0);

      /* count the pages and free memory */
      numPages++;
    }


LexPrintK_cancelled:

  if (eject)
    {
      LexEOP (out, 0);
      dbg_page_end (0);
    }

  if (sweepBuffer.buffer)
    free (sweepBuffer.buffer);

  FinishDataPipe (&dataPipe);

/*
  if (numPages == 0)
    {
      fprintf (stderr, "%s No pages printed!\n\n", ERROR_ENTRY);
      return 1;
    }
*/
  return 0;
}                               /* LexPrintK */


/***************************************************************************
 * returns the middle of a vertikal adjustment line to be printed
 ***************************************************************************/
static int
AdjustmentPrintMid (int cColumn, int nColumns, int width)
{
  return (width / (nColumns + 2)) * (cColumn + 1);
}

/***************************************************************************
 * print routine that prints a adjustment page to select the value
 ***************************************************************************/
int static
LexPrintKAdjust (FILE * in, FILE * out, TParameters * pParam)
{
  /* buffers to hold the print information */
  TSweepBuffer sweepBuffer = { 0 };
  /* number of pixels to move after a sweep has been printed */

  int cEmptyLines = 0;
  int cAdjustment = 1;

  /* some space values for nicer print outs */
  int emptyLinesOnTop = 1000;
  int emptyLinesInBetween = 30;

  /* experimentally determined line feed on B&W */
  cEmptyLines = BW_EMPTY_LINES_PRESET;


  Bool eject = False;

  /* Initialize the printer, load a page  */
  if (LexInit (out, pParam->termFileName))
    goto LexPrintKAdjust_cancelled;


  dbg_page_start (Z11_PAGE_HEIGHT, 0, False);

  dbg_empty_line (emptyLinesOnTop);
  cEmptyLines += 2 * emptyLinesOnTop;

  SweepBufferInit (&sweepBuffer, Z11_PAGE_WIDTH,
                   Z11_PAGE_WIDTH * BYTES_PER_COLUMN);
  sweepBuffer.bufpos = 0;


  while (cAdjustment <= MAX_ADJUST)
    {

      int cAdjLine;
      int cPixLine;
      int cAdj;

      for (cAdjLine = 0; cAdjLine < 2; ++cAdjLine)
        {
          /* run through all the pixel lines to check for the best offset */
          for (cPixLine = 0; cPixLine < BW_LINES; ++cPixLine)
            {
              int cAdjColumn;

              /* address the lines linearily through the cartridge */
              dbg_line_start (0);

              /* set line pixel to print */
              SetBWSweepParams (&sweepBuffer, cPixLine);

              /* flush the buffer of the last print out */
              if (cPixLine == 0)
                if (PrintBWSweep
                    (&sweepBuffer, out, BW_LINE_DRIVE, &cEmptyLines,
                     pParam->termFileName))
                  {
                    eject = pParam->ejectOnTerminate;
                    goto LexPrintKAdjust_cancelled;
                  }

              cAdj = cAdjustment;
              for (cAdjColumn = 0;
                   cAdj <= MAX_ADJUST && cAdjColumn < NEUTRAL_ADJUST;
                   ++cAdjColumn, ++cAdj)
                {
                  int cPixPos;
                  cPixPos =
                    AdjustmentPrintMid (cAdjColumn, NEUTRAL_ADJUST,
                                        Z11_PAGE_WIDTH) + GetAdjustment (cAdj,
                                                                         cPixLine);

                  dbg_set_bit (cPixPos);
                  SetSweepPixel (&sweepBuffer, 0, cPixPos);
                }
              dbg_line_end (0);
            }
        }

      /* quite complicated loop to draw the digits */
      for (cPixLine = 0; cPixLine < BW_LINES; ++cPixLine)
        {
          int digitLine;
          int cAdjColumn;
          dbg_line_start (0);

          /* set the line pixel to print */
          SetBWSweepParams (&sweepBuffer, cPixLine);

          /* make sure to flush the buffer */
          if (cPixLine == 0)
            if (PrintBWSweep (&sweepBuffer, out, BW_LINE_DRIVE, &cEmptyLines,
                              pParam->termFileName))
              {
                eject = pParam->ejectOnTerminate;
                goto LexPrintKAdjust_cancelled;
              }

          /* map the pixel line to the digit line and print only when in the range of the digits */
          digitLine = cPixLine - emptyLinesInBetween;
          if (digitLine >= 0 && digitLine < DIGIT_HEIGHT)
            {
              cAdj = cAdjustment;
              for (cAdjColumn = 0;
                   cAdj <= MAX_ADJUST && cAdjColumn < NEUTRAL_ADJUST;
                   ++cAdjColumn, ++cAdj)
                {
                  char szStr[4];        /* the string to hold the number */
                  int cPixPos;  /* counter, when going throght all the pixels of all digits */
                  int nDigits;  /* number of digits of the number */
                  int cDigit;   /* currently processed digit */

                  /* calculate the string that shall be printed */
                  sprintf (szStr, "%i", cAdj);
                  nDigits = strlen (szStr);

                  /* the left pixel of the first digit */
                  cPixPos =
                    AdjustmentPrintMid (cAdjColumn, NEUTRAL_ADJUST,
                                        Z11_PAGE_WIDTH) -
                    (DIGIT_WIDTH * nDigits) / 2;

                  /* iterate through the digits and set the pixels */
                  for (cDigit = 0; cDigit < nDigits; ++cDigit)
                    {
                      int cPixel;
                      for (cPixel = 0; cPixel < DIGIT_WIDTH;
                           ++cPixel, ++cPixPos)
                        {
                          if (GetDigitDot
                              (szStr[cDigit] - '0', cPixel, digitLine))
                            {
                              SetSweepPixel (&sweepBuffer, 0, cPixPos);
                              dbg_set_bit (cPixPos);
                            }
                        }
                    }
                }
            }
          dbg_line_end (0);
        }

      cAdjustment = cAdj;
    }
  /* final buffer flush */
  if (PrintBWSweep
      (&sweepBuffer, out, BW_LINE_DRIVE, 0, pParam->termFileName))
    {
      eject = pParam->ejectOnTerminate;
      goto LexPrintKAdjust_cancelled;
    }

  /* eat any remaining whitespace as long there is data in the pipe
     so process will not hang */
  if (LexEOP (out, pParam->termFileName))
    {
      eject = pParam->ejectOnTerminate;
      goto LexPrintKAdjust_cancelled;
    }

  dbg_page_end (0);

LexPrintKAdjust_cancelled:

  if (eject)
    {
      LexEOP (out, 0);
      dbg_page_end (0);
    }

  if (sweepBuffer.buffer)
    free (sweepBuffer.buffer);

  while (getc (in) != EOF);

  return 0;
}

/***************************************************************************
 * sweep trough a string until a delimiter is reached and return
 * the contained integer value
 * if the value would be bigger than maxValue the reading stops
 * if a non-digit is read the reading stops
 * if stoppedAt != 0 it contains the pointer where the read stopped
 ***************************************************************************/
static int
GetPositiveValue (char *line, int maxValue, char delimiter, char **stoppedAt)
{
  char c;
  int value = 0;
  while ((c = *line) != 0 && c != delimiter)
    {
      int v = c - '0';
      if (v >= 0 && v <= 9 && (value + 1) < (INT_MAX / 10)
          && (value * 10) <= (maxValue - v))
        {
          value = value * 10 + v;
          ++line;
        }
      else
        break;
    }
  if (stoppedAt)
    *stoppedAt = line;
  return value;
}


/***************************************************************************
 * returns the start address of a option string
 ***************************************************************************/
static char *
DecodeOption (char *option, char *line)
{
  int len = strlen (option);
  if (!strncmp (option, line, len))
    return &line[len];
  else
    return 0;
}


#define OPT_BITCMYK     0
#define OPT_BITRGB      1
#define OPT_BIT         2
#define OPT_SIZE        3
#define OPT_ADJUST      4
#define OPT_HELP        5
#define OPT_VERSION     6
#define OPT_TERMINATE   7
#define OPT_TERMINATEEJECT   8
#define OPT_UNKNOWN     9
#define NUM_OPT OPT_UNKNOWN


/***************************************************************************
 * returns the type of a option and sets the pointer to a option string
 ***************************************************************************/
static int
GetOptionType (char *arg, char **pResult)
{
  int i;
  static char *optStrings[] = { "--bitcmyk=",
    "--bitrgb=",
    "--bit=",
    "--size=",
    "--adjust=",
    "--help",
    "--version",
    "-c",
    "-r",
    "-b",
    "-s",
    "-a",
    "-h",
    "-v",
    "--terminate=",
    "--terminate-eject=",
    ""
  };

  static int optValue[] = { OPT_BITCMYK,
    OPT_BITRGB,
    OPT_BIT,
    OPT_SIZE,
    OPT_ADJUST,
    OPT_HELP,
    OPT_VERSION,
    OPT_BITCMYK,
    OPT_BITRGB,
    OPT_BIT,
    OPT_SIZE,
    OPT_ADJUST,
    OPT_HELP,
    OPT_VERSION,
    OPT_TERMINATE,
    OPT_TERMINATEEJECT,
    OPT_UNKNOWN
  };


  for (i = 0; *optStrings[i]; ++i)
    {
      if ((*pResult = DecodeOption (optStrings[i], arg)) != 0)
        {
          return optValue[i];
        }
    }
  return OPT_UNKNOWN;
}

/***************************************************************************
 * extract the integers out of string
 * <sizex>x<sizey> in the form "nnnxnnn"
 * return True on success
 ***************************************************************************/
static Bool
DecodeSize (char *line, int *width, int *height)
{
  char *stoppedAt;
  int w, h;
  w = GetPositiveValue (line, INT_MAX / 10, 'x', &stoppedAt);
  if (!w || *stoppedAt != 'x')
    return False;
  h = GetPositiveValue (stoppedAt + 1, INT_MAX / 10, '\0', &stoppedAt);
  if (!h || *stoppedAt != '\0')
    return False;
  *width = w;
  *height = h;
  return True;
}

/***************************************************************************
 * decode the dithering type
 * return True on success
 ***************************************************************************/
static Bool
DecodeColorDithering (char *line, int *dithering)
{
  static char *dString[] =
    { "1.5x1", "2x1", "1.5x2", "2x2", "3x2", "1x1", "1x1--gray-scale", "" };
  static int dRes[] = { 0, 1, 2, 3, 4, RANDOM_DITHERING, GRAY_DITHERING };
  int i;
  for (i = 0; *dString[i]; ++i)
    {
      if (!strcmp (line, dString[i]))
        {
          *dithering = dRes[i];
          return True;
        }
    }
  return False;
}


/***************************************************************************
 * decode the dithering type
 * return True on success
 ***************************************************************************/
static Bool
DecodeBlackDithering (char *line, int *dithering)
{
  static char *dString[] = { "1x1", "1.5x1", "2x1", "2x2", "3x2", "" };
  static int dRes[] = { 0, 1, 2, 3, 4 };
  int i;
  for (i = 0; *dString[i]; ++i)
    {
      if (!strcmp (line, dString[i]))
        {
          *dithering = dRes[i];
          return True;
        }
    }
  return False;
}


/***************************************************************************
 * determine the name of the file, where the PID is written for the
 * driver instances that shall terminate
 ***************************************************************************/
static Bool
DecodeTerminateFile (char *line, char *termFileName, int size)
{
  if (strlen (line) >= size)
    return False;
  else
    strcpy (termFileName, line);
  return True;
}


/***************************************************************************
 * decode the adjustment type
 * return True on success
 ***************************************************************************/
static Bool
DecodeAdjustment (char *line, int *adjustment)
{
  char *stoppedAt;
  int a;
  a = GetPositiveValue (line, MAX_ADJUST, '\0', &stoppedAt);
  if (a && *stoppedAt == '\0')
    {
      *adjustment = a;
      return True;
    }
  if (a == 0 && strcmp (stoppedAt, "test") == 0)
    {
      *adjustment = TESTPAGE_ADJUST;
      return True;
    }
  return False;
}

/***************************************************************************
 * strings to display for help etc.
 ***************************************************************************/
static char VersionInfo[] =
  "\n"
  "BITCYMK/BIT to Lexmark Z11 stream converter\n"
  "Version %s\n"
  "\n"
  "(C) 2000-2004 Christian Kornblum, Erik Bernhardson,\n"
  "              Som-Som, Mihael Hategan, Ullrich Sigwanz\n"
  "              and Till Kamppeter\n" "\n";

static char ShortHelp[] = " --help, -h   for more information.\n" "\n";

static char LongHelp[] =
  "\n"
  " -c<dithering>\n"
  " --bitcmyk=<dithering>     dithering for a color bitcmyk-bitmap\n"
  "                           The driver will convert 1 pixel into the\n"
  "                           given number of printer dots.\n"
  "                           allowed values are:\n"
  "                           1.5x1 , 2x1 , 1.5x2 , 2x2 , 3x2\n"
  "                             and\n"
  "                           1x1 (which is in for backward compatibilty\n"
  "                           to print 600dpi x 600dpi BITCMYKs)\n"
  "\n"
  "                           1x1--gray-scale will print BITCMYKs\n"
  "                           to a black cartridge (less crisp look)\n"
  "\n"
  "                           examples: -c3x2 , --bitcmyk=1.5x2\n"
  "\n"
  " -b<dithering>\n"
  " --bit=<dithering>         dithering for a monochrome bit-bitmap\n"
  "                           The driver will convert 1 pixel into the\n"
  "                           given number of printer dots.\n"
  "                           allowed values are:\n"
  "                           1x1 , 1.5x1 , 2x1 , 2x2 , 3x2\n"
  "\n"
  "                           examples: -b1.5x1 , --bit=2x2\n"
  "\n"
  "\n"
  " -s<sizeX>x<sizeY>\n"
  " --size=<sizeX>x<sizeY>    size of the bitmap\n"
  "                           For bit-bitmaps the SizeX-number will be extended to \n"
  "                           the next multiple of 8.\n"
  "                           For bitcmyk-bitmaps the Sizex-number will be extended\n"
  "                           to the next multiple of 2.\n"
  "\n"
  "                           examples: -s4958x7017, --size=4958x7017\n"
  "\n"
  " -a<1-64|test>             give an adjustment value to correct the skew of\n"
  " --adjust=<1-64|test>      the black cartridge. With the option value\n"
  "                           \"test\" you can print samples of lines.\n"
  "                           Choose the value of the less skewed line for\n"
  "                           further printouts (black ink only).\n"
  "\n"
  "                           examples --adjust=test, -a22\n"
  "\n"
  " --terminate=<filename>    name of a file that holds the driver-PID that \n"
  "                           shall be stopped. This gives the chance to stop\n"
  "                           print outs that are wrong and would produce a \n"
  "                           heap of spoiled paper.\n"
  "\n"
  " --terminate-eject=<filename> same as --terminate, but will eject the \n"
  "                           last fed page \n"
  "\n"
  " -v\n"
  " --version                 print version string to stderr\n"
  "\n"
  " -h\n"
  " --help                    print this help\n"
  "\n"
  " feedback, suggestions, bug reports to:\n"
  " pohl.sigwanz@freesurf.ch\n" "\n";


/***************************************************************************
 * Print Help Info and return with exitCode
 ***************************************************************************/
static void
PrintHelp (int exitCode)
{
  fprintf (stderr, VersionInfo, VERSION);
  if (exitCode)
    fprintf (stderr, "%s", ShortHelp);
  else
    fprintf (stderr, "%s", LongHelp);
  exit (exitCode);
}

/***************************************************************************
 *  print version
 ***************************************************************************/
static void
PrintVersion (void)
{
  fprintf (stderr, "%s\n", VERSION);
  exit (0);
}

/***************************************************************************
 * give information about a invalid invokation , print help and exit with
 * code 1
 ***************************************************************************/
static void
InvalidInvokation (char *msg)
{
  fprintf (stderr, "%s invalid invokation parameter %s\n", ERROR_ENTRY, msg);
  PrintHelp (1);
}

/***************************************************************************
 * give information about conflicting invokation , print help and exit with
 * code 1
 ***************************************************************************/
static void
ConflictingInvokation (char *msg1, char *msg2)
{
  fprintf (stderr, "%s conflicting invokation parameters %s %s\n",
           ERROR_ENTRY, msg1, msg2);
  PrintHelp (1);
}

/***************************************************************************
 * give information about duplicate optione , print help and exit with
 * code 1
 ***************************************************************************/
static void
DuplicateInvokation (char *msg)
{
  fprintf (stderr, "%s invokation parameter %s... occurs more than once\n",
           ERROR_ENTRY, msg);
  PrintHelp (1);
}

/***************************************************************************
 * give information about missing optione , print help and exit with
 * code 1
 ***************************************************************************/
static void
IncompleteInvokation (char *msg)
{
  fprintf (stderr, "%s invokation is incomplete\nmissing:\n%s", ERROR_ENTRY,
           msg);
  PrintHelp (1);
}

/***************************************************************************
 * give information about discardable options
 ***************************************************************************/
static void
IgnoringParameter (char *szParam, char *msg)
{
  fprintf (stderr, "%s warning: ignoring -%s- parameter: %s\n", ERROR_ENTRY,
           szParam, msg);
}



/***************************************************************************
 * set the parameters appropriate according to an option like style
 ***************************************************************************/
static Bool
GetArgs (int argc, char *argv[], TParameters * pParam)
{

  Bool isAnyParam = False;
  Bool isDither = False;
  Bool isSize = False;
  Bool isAdjust = False;
  char szSize[] = " --size=<sizeX>x<sizeX>\n";
  char szDither[] = " --bitcmyk=<dithering> | --bit=<dithering>\n";
  char szSizeAndDither[] =
    " --size=<sizeX>x<sizeX>  --bitcmyk=<dithering> | --bit=<dithering>\n";

  int isOpt[NUM_OPT] = { 0 };

  int i;
  char *restOfLine;

  for (i = 1; i < argc; ++i)
    {

      int type = GetOptionType (argv[i], &restOfLine);

      if (type == OPT_UNKNOWN)
        InvalidInvokation (argv[i]);

      else if (isOpt[type])
        {
          *restOfLine = 0;
          DuplicateInvokation (argv[i]);
        }

      switch (type)
        {

          /* request for version information */
        case OPT_VERSION:
          isAnyParam = True;
          if (!*restOfLine)
            PrintVersion ();
          else
            InvalidInvokation (argv[i]);
          break;

          /* request for a verbose help page */
        case OPT_HELP:
          isAnyParam = True;
          if (!*restOfLine)
            PrintHelp (0);
          else
            InvalidInvokation (argv[i]);
          break;

          /* set the dithering for cmyk */
        case OPT_BITCMYK:
          isAnyParam = isDither = True;
          {
            if (isOpt[type])
              {
                *restOfLine = 0;
                DuplicateInvokation (argv[i]);
              }
            else if (isOpt[OPT_BITRGB])
              ConflictingInvokation (argv[isOpt[OPT_BITRGB]], argv[i]);
            else if (isOpt[OPT_BIT])
              ConflictingInvokation (argv[isOpt[OPT_BIT]], argv[i]);
            else if (!DecodeColorDithering (restOfLine, &pParam->dithering))
              InvalidInvokation (argv[i]);

            /* select the proper print mode and dithering */
            if (pParam->dithering == GRAY_DITHERING)
              {
                pParam->dithering = 0;
                pParam->printMode = PRINT_GRAYCMYK;
              }
            else
              {
                pParam->printMode = PRINT_BITCMYK;
              }
          }
          break;

          /* set the dithering for rgb */
        case OPT_BITRGB:
          isAnyParam = isDither = True;
          {
            if (isOpt[type])
              {
                *restOfLine = 0;
                DuplicateInvokation (argv[i]);
              }
            else if (isOpt[OPT_BITCMYK])
              ConflictingInvokation (argv[isOpt[OPT_BITCMYK]], argv[i]);
            else if (isOpt[OPT_BIT])
              ConflictingInvokation (argv[isOpt[OPT_BIT]], argv[i]);
            else if (!DecodeColorDithering (restOfLine, &pParam->dithering))
              InvalidInvokation (argv[i]);

            /* select the proper print mode and dithering */
            if (pParam->dithering == GRAY_DITHERING)
              {
                pParam->dithering = 0;
                pParam->printMode = PRINT_GRAYRGB;
              }
            else
              {
                pParam->printMode = PRINT_BITRGB;
              }
          }
          break;

          /* set the dithering for a black and white bit-device */
        case OPT_BIT:
          isAnyParam = isDither = True;
          {
            if (isOpt[type])
              {
                *restOfLine = 0;
                DuplicateInvokation (argv[i]);
              }
            else if (isOpt[OPT_BITCMYK])
              ConflictingInvokation (argv[isOpt[OPT_BITCMYK]], argv[i]);
            else if (isOpt[OPT_BITRGB])
              ConflictingInvokation (argv[isOpt[OPT_BITRGB]], argv[i]);
            else if (!DecodeBlackDithering (restOfLine, &pParam->dithering))
              InvalidInvokation (argv[i]);

            pParam->printMode = PRINT_BIT;
          }
          break;

          /* read the size information */
        case OPT_SIZE:
          isAnyParam = isSize = True;
          if (isOpt[type])
            {
              *restOfLine = 0;
              DuplicateInvokation (argv[i]);
            }
          else
            if (!DecodeSize
                (restOfLine, &pParam->pageWidth, &pParam->pageHeight))
            InvalidInvokation (argv[i]);
          break;

          /* check the skew adjustment */
        case OPT_ADJUST:
          isAnyParam = isAdjust = True;
          if (isOpt[type])
            {
              *restOfLine = 0;
              DuplicateInvokation (argv[i]);
            }
          else if (!DecodeAdjustment (restOfLine, &pParam->adjustment))
            InvalidInvokation (argv[i]);
          break;

        case OPT_TERMINATEEJECT:
          pParam->ejectOnTerminate = True;
          /* fall through */

        case OPT_TERMINATE:
          if (!DecodeTerminateFile (restOfLine, pParam->termFileName, 256))
            InvalidInvokation (argv[i]);
          break;

        default:
          InvalidInvokation (argv[i]);
          break;
        }
      isOpt[type] = i;
    }

  /* when printing test page, disregard the dithering */
  if (pParam->adjustment == TESTPAGE_ADJUST)
    {
      /* size is always DIN A 4 for skew test */
      if (isSize)
        IgnoringParameter ("size", argv[isOpt[OPT_SIZE]]);

      /* ignore all dithering information */
      if (isOpt[OPT_BITRGB])
        IgnoringParameter ("BITRGB", argv[isOpt[OPT_BITRGB]]);
      else if (isOpt[OPT_BITCMYK])
        IgnoringParameter ("BITCMYK", argv[isOpt[OPT_BITCMYK]]);
      else if (isOpt[OPT_BIT])
        IgnoringParameter ("BIT", argv[isOpt[OPT_BIT]]);

      return True;
    }

  /* if size or dither is missing issue an error */
  if (isDither != isSize)
    {
      IncompleteInvokation (!isSize ? szSize : szDither);
      return False;
    }

  /* if only adjust is passed, issue an error */
  if (!(isSize || isDither) && isAdjust)
    {
      IncompleteInvokation (szSizeAndDither);
      return False;
    }

  /* otherwise show a short help */
  if (!isAnyParam)
    {
      PrintHelp (1);
      return False;
    }

  /* adjust is ignored when no bit device is  selected */
  if (isAdjust && ColorPrint (pParam->printMode))
    {
      IgnoringParameter ("adjust", argv[isOpt[OPT_ADJUST]]);
    }

  return True;
}

/***************************************************************************
 * The main program. Sets input and output streams.
 ***************************************************************************/
int
main (int argc, char *argv[])
{
  FILE *InputFile;
  FILE *OutPutFile;

  int result = 0;               /* return value of print routine */

  /* assign the defaults first */
  TParameters param = { 0 };

  param.printMode = PRINT_BITCMYK;

  param.pageWidth = (4958 / 2);
  param.pageHeight = (7017 / 2);
  param.dithering = 2;
  param.adjustment = NEUTRAL_ADJUST;


  /* assign in and out */
  InputFile = stdin;
  OutPutFile = stdout;


  if (!GetArgs (argc, argv, &param))
    return 1;

  /* assign in and out */
  InputFile = stdin;
  OutPutFile = stdout;

  /* print the file */
  if (param.adjustment == 0)
    result = LexPrintKAdjust (InputFile, OutPutFile, &param);
  else if (BlackPrint (param.printMode))
    result = LexPrintK (InputFile, OutPutFile, &param);
  else
    result = LexPrintCMYK (InputFile, OutPutFile, &param);

  /* close the pipes */
  fclose (OutPutFile);
  fclose (InputFile);

  return result;
}
